/*
* @license Apache-2.0
*
* Copyright (c) 2022 The Stdlib Authors.
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*    http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/

// TypeScript Version: 2.0

/// <reference types="@stdlib/types"/>

import { Collection } from '@stdlib/types/object';
import { ndarray } from '@stdlib/types/ndarray';

/**
* Callback invoked for each array element.
*
* @returns result
*/
type Nullary = () => any;

/**
* Callback invoked for each array element.
*
* @param value - array element
* @returns result
*/
type Unary = ( value: any ) => any;

/**
* Callback invoked for each array element.
*
* @param value - array element
* @param index - element index
* @returns result
*/
type Binary = ( value: any, index: number ) => any;

/**
* Callback invoked for each array element.
*
* @param value - array element
* @param index - element index
* @param arr - array
* @returns result
*/
type Ternary = ( value: any, index: number, arr: ndarray ) => any;

/**
* Callback invoked for each array element.
*
* @param value - array element
* @param index - element index
* @param arr - array
* @returns result
*/
type ArrayTernary = ( value: any, index: number, arr: Collection ) => any;

/**
* Callback invoked for each array element.
*
* @param value - array element
* @param index - element index
* @param arr - array
* @returns result
*/
type Mapper = Nullary | Unary | Binary | Ternary;

/**
* Callback invoked for each array element.
*
* @param value - array element
* @param index - element index
* @param arr - array
* @returns result
*/
type ArrayMapper = Nullary | Unary | Binary | ArrayTernary;

/**
* Function applied against an accumulator.
*
* @returns accumulator value
*/
type NullaryReducer = () => void;

/**
* Function applied against an accumulator.
*
* @param accumulator - accumulated value
* @returns accumulator value
*/
type UnaryReducer = ( accumulator: any ) => any;

/**
* Function applied against an accumulator.
*
* @param accumulator - accumulated value
* @param value - array element
* @returns accumulator value
*/
type BinaryReducer = ( accumulator: any, value: any ) => any;

/**
* Function applied against an accumulator.
*
* @param accumulator - accumulated value
* @param value - array element
* @param index - element index
* @returns accumulator value
*/
type TernaryReducer = ( accumulator: any, value: any, index: number ) => any;

/**
* Function applied against an accumulator.
*
* @param accumulator - accumulated value
* @param value - array element
* @param index - element index
* @param arr - array
* @returns accumulator value
*/
type QuaternaryReducer = ( accumulator: any, value: any, index: number, arr: ndarray ) => any; // tslint:disable-line:max-line-length

/**
* Function applied against an accumulator.
*
* @param accumulator - accumulated value
* @param value - array element
* @param index - element index
* @param arr - array
* @returns accumulator value
*/
type ArrayQuaternaryReducer = ( accumulator: any, value: any, index: number, arr: Collection ) => any; // tslint:disable-line:max-line-length

/**
* Function applied against an accumulator.
*
* @param accumulator - accumulated value
* @param value - array element
* @param index - element index
* @param arr - array
* @returns accumulator value
*/
type Reducer = NullaryReducer | UnaryReducer | BinaryReducer | TernaryReducer | QuaternaryReducer; // tslint:disable-line:max-line-length

/**
* Function applied against an accumulator.
*
* @param accumulator - accumulated value
* @param value - array element
* @param index - element index
* @param arr - array
* @returns accumulator value
*/
type ArrayReducer = NullaryReducer | UnaryReducer | BinaryReducer | TernaryReducer | ArrayQuaternaryReducer; // tslint:disable-line:max-line-length

/**
* Performs a map-reduce operation for each element in an array while iterating from right to left and returns the accumulated result.
*
* ## Notes
*
* -   The mapping function is provided four arguments:
*
*     -   **value**: array element.
*     -   **index**: element index.
*     -   **arr**: input array.
*
* -   The reducing function is provided four arguments:
*
*     -   **accumulator**: accumulated value.
*     -   **value**: result after applying a mapping function to the current array element.
*     -   **index**: element index.
*     -   **arr**: input array.
*
* -   If provided an empty array, the function returns the initial value.
*
* @param arr - input array
* @param initial - initial value
* @param mapper - mapping function
* @param reducer - reducing function
* @param thisArg - reducing function execution context
* @returns accumulated result
*
* @example
* var naryFunction = require( `@stdlib/utils/nary-function` );
* var abs = require( `@stdlib/math/base/special/abs` );
* var add = require( `@stdlib/math/base/ops/add` );
* var array = require( `@stdlib/ndarray/array` );
*
* var opts = {
*     'dtype': 'generic'
* };
* var arr = array( [ [ -1, -2, -3 ], [ -4, -5, -6 ] ], opts );
*
* var out = mapReduceRight( arr, 0, naryFunction( abs, 2 ), naryFunction( add, 2 ) );
* // returns 21
*/
declare function mapReduceRight( arr: ndarray, initial: any, mapper: Mapper, reducer: Reducer, thisArg?: any ): any; // tslint:disable-line:max-line-length

/**
* Performs a map-reduce operation for each element in an array while iterating from right to left and returns the accumulated result.
*
* ## Notes
*
* -   The mapping function is provided four arguments:
*
*     -   **value**: array element.
*     -   **index**: element index.
*     -   **arr**: input array.
*
* -   The reducing function is provided four arguments:
*
*     -   **accumulator**: accumulated value.
*     -   **value**: result after applying a mapping function to the current array element.
*     -   **index**: element index.
*     -   **arr**: input array.
*
* -   If provided an empty array, the function returns the initial value.
*
* @param arr - input array
* @param initial - initial value
* @param mapper - mapping function
* @param reducer - reducing function
* @param thisArg - reducing function execution context
* @returns accumulated result
*
* @example
* function square( value ) {
*     return value * value;
* }
*
* function sum( acc, value ) {
*     return acc + value;
* }
*
* var arr = [ 1, 2, 3, 4 ];
*
* var out = mapReduceRight( arr, 0, square, sum );
* // returns 30
*/
declare function mapReduceRight( arr: Collection, initial: any, mapper: ArrayMapper, reducer: ArrayReducer, thisArg?: any ): any; // tslint:disable-line:max-line-length


// EXPORTS //

export = mapReduceRight;
